﻿-------------------------------------------------------------------------------
-- Addon - Combo-Points - Init
-------------------------------------------------------------------------------
-- Version
local MAJOR = "Avion2_CPE"
local MINOR = tonumber(("8"):match("%d+")) or 0

-- Dependencies
if not Avion2Core 				then error(MAJOR .. " requires Avion2.") end
if not LibStub("AceEvent-3.0")	then error(MAJOR .. " requires AceEvent-3.0.") end

-- Create
Avion2_CPE 			= Avion2Core:NewModule(MAJOR, "AceEvent-3.0")
Avion2_CPE.major 	= MAJOR
Avion2_CPE.minor 	= MINOR


-------------------------------------------------------------------------------
-- Addon - Combo-Points - Variables
-------------------------------------------------------------------------------
-- Translation
local L 	= LibStub("AceLocale-3.0"):GetLocale(MAJOR)
local coreL	= LibStub("AceLocale-3.0"):GetLocale(Avion2Core.major)

-- Save settings for new trigger
local newSettings = {
	comboPoints	= 0,
	energy 		= 0,
}

-- Local functions
local ComboPointRelay, EnergyRelay, ComboPointUpdate, EnergyUpdate, ComboPointEnergyUpdate


-------------------------------------------------------------------------------
-- Addon - Combo-Points - Register with core
-------------------------------------------------------------------------------
-- Register SavedVariables
function Avion2_CPE:RegisterModuleDB()
	-- Register own saved variables DB
	self.db = Avion2Core.db:RegisterNamespace(Avion2_CPE.major, {
		profile = {
			combinations 				= {},
			onlyHighestComboPoints 		= false,
			onlyHighestEnergy 			= false,
			
			energy 						= {},
			onlyOneActiveEnergy 		= false,
			
			comboPoint					= {},
			onlyOneActiveComboPoint		= false,
			
			lastMinor 					= MINOR,
		},
	})
	
	-- Last combo-points
	self.lastComboPoints = 0
	
	-- Track last energy value
	self.lastUnitPower = 0
	
	-- Updateto to new rev, data converters should go here
	self.db.profile.lastMinor = Avion2_CPE.minor
end


-- (Un)register (un)needed events
function Avion2_CPE:UpdateEventRegistration()
	-- Combo-Points events
	if (self:IsEnabled() and next(self.db.profile.combinations)) then
		self:RegisterEvent("UNIT_ENERGY", EnergyRelay)
		self:RegisterEvent("PLAYER_COMBO_POINTS", ComboPointRelay)
	else
		self:UnregisterEvent("UNIT_ENERGY")
		self:UnregisterEvent("PLAYER_COMBO_POINTS")
	end
end


-- Update animation (convert animationFunc: string name to function pointer)
function Avion2_CPE:LoadAnimationTables()
	for _, info in pairs(self.db.profile.combinations) do
		Avion2Core:LoadAnimationTables(info.animationList)
	end
	
	for _, data in pairs(self.db.profile.energy) do
		Avion2Core:LoadAnimationTables(data.animationList)
	end
	
	for _, data in pairs(self.db.profile.comboPoint) do
		Avion2Core:LoadAnimationTables(data.animationList)
	end
end


-------------------------------------------------------------------------------
-- Addon - Combo-Points - Options
-------------------------------------------------------------------------------
-- Fetch animationTable from animationID (as suplied by eg. Menu_Reuqest)
function Avion2_CPE:GetAnimationTable(animationID)
	success, _, arg1, arg2, arg3, arg4 = self:Deserialize(animationID)
	
	if (success) then
		if (arg1 == "Energy") then
			return self.db.profile.energy[arg2].animationList[arg3]
		elseif (arg1 == "ComboPoint") then
			return self.db.profile.comboPoint[arg2].animationList[arg3]
		elseif (arg1 == "Combination") then
			for key, info in pairs(self.db.profile.combinations) do
				if (info.comboPoints == arg2 and info.energy == arg3) then
					return self.db.profile.combinations[key].animationList[arg4]
				end
			end
		end
	else
		return false
	end
end


-- Sorting function, will sort table by (1) comb-points DESC, (2) energy DESC
local function comp (val1, val2)
	-- Sort by combo-points first
	if (val1.comboPoints < val2.comboPoints) then
		return false -- not true
		
	-- If equal, ...
	elseif (val1.comboPoints == val2.comboPoints) then
		-- ... sort by energie
		if (val1.energy < val2.energy) then
			return false -- not true
		else
			return  true -- not false
		end
	else
		return true -- not false
	end
end


-- Build Optionstable
function Avion2_CPE:UpdateOptionTable(args)
	-- Update options when enabled and initialized
	if (self.db and self:IsEnabled()) then
		local serializedInfo
		
		-- Final options (ENERGY HERE)
		args[self.major.."_Energy"] = {
			type = 'group',
			name = L["ThresholdsEnergy"],
			desc = L["ThresholdsEnergy_Desc"],
			args = {
				new = {
					order = 1,
					type = 'input',
					name = L["NewThreshold"],
					desc = L["NewThresholdEnergy_Desc"],
					usage = L["NewThresholdEnergy_Usage"],
					pattern = "(%d)",
					get = function()
						return ""
					end,
					set = function(info, val)
						if (not self.db.profile.energy[tonumber(val)]) then
							self.db.profile.energy[tonumber(val)] = {
								animationList 	= {},
								cancelAnimation	= 0,
							}
							
							self:UpdateEventRegistration()
							Avion2Core:UpdateModuleOptionTable(self)
						end
					end,
				},
				onlyOneActiveEnergy = {
					order = 2,
					type = 'toggle',
					name = L["OnlyOneActive"],
					desc = L["OnlyOneActive_Desc"],
					get = function()
						return self.db.profile.onlyOneActiveEnergy
					end,
					set = function()
						self.db.profile.onlyOneActiveEnergy = not self.db.profile.onlyOneActiveEnergy
					end,
				},
			},
		}		
		-- Options for each energy threshold
		for index, data in pairs(self.db.profile.energy) do
			args[self.major.."_Energy"].args[string.format("Threshold_%d", index)] = {
				type = 'group',
				name = string.format(L["Threshold"], index),
				desc = string.format(L["ThresholdEnergy_Desc"], index),
				args = { 
					list = {
						order = 1,
						type = 'group',
						name = coreL["AnimationList"],
						desc = coreL["AnimationList_Desc"],
						args = {
							new = Avion2Core:GetDefaultNewAnimationOption(data.animationList),
						},
					},
					cancelAnimation = {
						order = 2,
						type = 'select',
						name = coreL["CancelAnimation"],
						desc = coreL["CancelAnimation_Desc"],
						values = {
							[0] = coreL["DontCancel"],
							[1] = coreL["FadeOut"],
							[2] = coreL["Cancel"],
						},
						get = function()
							return data.cancelAnimation
						end,
						set = function(info, val)
							data.cancelAnimation = val
						end,
					},
					delete = {
						order = 3,
						type = 'execute',
						name = L["DeleteThreshold"],
						desc = L["DeleteThreshold_Desc"],
						func = function()
							Avion2Core:StopAnimationList(string.format("Avion2_CPEThreshold_%d", index), data.animationList, false)
							self.db.profile.energy[index] = nil
							
							self:UpdateEventRegistration()
							Avion2Core:UpdateModuleOptionTable(self)
						end,
					},
				},
			}
			-- Add options for each animation
			for animationListID, _ in pairs(data.animationList) do
				-- Combine all info to find this animation in module db in ONE string
				serializedInfo = self:Serialize(Avion2_CPE.major, "Energy", index, animationListID)
				
				-- Update option table
				args[self.major.."_Energy"].args[string.format("Threshold_%d", index)].args.list.args[string.format("Animation_%d", animationListID)] = Avion2Core:GetDefaultAnimationOption(data.animationList, animationListID, serializedInfo)
			end
		end	

		-- Final options (COMBOPOINTS HERE)
		args[self.major.."_ComboPoint"] = {
			type = 'group',
			name = L["ThresholdsComboPoint"],
			desc = L["ThresholdsComboPoint_Desc"],
			args = {
				new = {
					order = 1,
					type = 'input',
					name = L["NewThreshold"],
					desc = L["NewThresholdComboPoint_Desc"],
					usage = L["NewThresholdComboPoint_Usage"],
					pattern = "(%d)",
					get = function()
						return ""
					end,
					set = function(info, val)
						if (not self.db.profile.comboPoint[tonumber(val)]) then
							self.db.profile.comboPoint[tonumber(val)] = {
								animationList 	= {},
								cancelAnimation	= 0,
							}
							
							self:UpdateEventRegistration()
							Avion2Core:UpdateModuleOptionTable(self)
						end
					end,
				},
				onlyOneActiveComboPoint = {
					order = 2,
					type = 'toggle',
					name = L["OnlyOneActive"],
					desc = L["OnlyOneActive_Desc"],
					get = function()
						return self.db.profile.onlyOneActiveComboPoint
					end,
					set = function()
						self.db.profile.onlyOneActiveComboPoint = not self.db.profile.onlyOneActiveComboPoint
					end,
				},
			},
		}		
		-- Options for each compo-point
		for index, data in pairs(self.db.profile.comboPoint) do
			args[self.major.."_ComboPoint"].args[string.format("Threshold_%d", index)] = {
				type = 'group',
				name = string.format(L["Threshold"], index),
				desc = string.format(L["ThresholdComboPoint_Desc"], index),
				args = { 
					list = {
						order = 1,
						type = 'group',
						name = coreL["AnimationList"],
						desc = coreL["AnimationList_Desc"],
						args = {
							new = Avion2Core:GetDefaultNewAnimationOption(data.animationList),
						},
					},
					cancelAnimation = {
						order = 2,
						type = 'select',
						name = coreL["CancelAnimation"],
						desc = coreL["CancelAnimation_Desc"],
						values = {
							[0] = coreL["DontCancel"],
							[1] = coreL["FadeOut"],
							[2] = coreL["Cancel"],
						},
						get = function()
							return data.cancelAnimation
						end,
						set = function(info, val)
							data.cancelAnimation = val
						end,
					},
					delete = {
						order = 3,
						type = 'execute',
						name = L["DeleteThreshold"],
						desc = L["DeleteThreshold_Desc"],
						func = function()
							Avion2Core:StopAnimationList(string.format("Avion2_ComboPointThreshold_%d", index), data.animationList, false)
							self.db.profile.comboPoint[index] = nil
							
							self:UpdateEventRegistration()
							Avion2Core:UpdateModuleOptionTable(self)
						end,
					},
				},
			}
			-- Add options for each animation
			for animationListID, _ in pairs(data.animationList) do
				-- Combine all info to find this animation in module db in ONE string
				serializedInfo = self:Serialize(Avion2_CPE.major, "ComboPoint", index, animationListID)
				
				-- Update option table
				args[self.major.."_ComboPoint"].args[string.format("Threshold_%d", index)].args.list.args[string.format("Animation_%d", animationListID)] = Avion2Core:GetDefaultAnimationOption(data.animationList, animationListID, serializedInfo)
			end
		end	
		
		-- Final options (COMBINATIONS HERE)
		args[self.major .. "_Combination"] = {
			type = 'group',
			name = L["ComboPointsEnergy"],
			desc = L["ComboPointsEnergy_Desc"],
			args = {
				new = {
					type = 'group',
					order = 1,
					name = L["NewComboPointsEnergy"],
					desc = L["NewComboPointsEnergy_Desc"],
					args = {
						comboPoints = {
							order = 1,
							type = 'input',
							name = L["NewComboPoints"],
							desc = L["NewComboPoints_Desc"],
							usage = L["NewComboPoints_Usage"],
							pattern = "(%d)",
							get = function()
								return tostring(newSettings.comboPoints)
							end,
							set = function(info, val)
								newSettings.comboPoints = tonumber(val)
							end,
						},
						energy = {
							order = 2,
							type = 'input',
							name = L["NewEnergy"],
							desc = L["NewEnergy_Desc"],
							usage = L["NewEnergy_Usage"],
							pattern = "(%d)",
							get = function()
								return tostring(newSettings.energy)
							end,
							set = function(info, val)
								newSettings.energy = tonumber(val)
							end,
						},
						addMe = {
							order = 3,
							type = 'execute',
							name = L["AddMe"],
							desc = L["AddMe_Desc"],
							func = function(info)
								if (newSettings.comboPoints >= 0 and newSettings.energy >= 0 and newSettings.comboPoints <= 5 and newSettings.energy <= 110) then
									-- Igor, did we find it??? Igor?
									local found = false
										
									-- Search it
									for _, info in pairs(self.db.profile.combinations) do
										if (info.comboPoints == newSettings.comboPoints and info.energy == newSettings.energy) then
											found = true
											break
										end
									end
									
									-- Add if there isn't allready one
									if (not found) then
										-- Add
										table.insert(self.db.profile.combinations, {comboPoints = newSettings.comboPoints, energy = newSettings.energy, cancelAnimation = 0, animationList = {}})
										
										-- Sort
										table.sort(self.db.profile.combinations, comp)
										
										-- Stop all running animation (due to resorting of animation table) NOTE: Only <need> to stop this modules own animaions, .. well.
										--!!!LibAvion:StopAllAnimations(false)
										
										-- Update registration of events and (oh shit) we need to recreate options table, since we resorted everything -.-
										self:UpdateEventRegistration()
										Avion2Core:UpdateModuleOptionTable(self)
									end
								end
							end
						},
					},
				},
				onlyHighestComboPoints = {
					type = 'toggle',
					name = L["OnlyHighestComboPoints"],
					desc = L["OnlyHighestComboPoints_Desc"],
					get = function()
						return self.db.profile.onlyHighestComboPoints
					end,
					set = function()
						self.db.profile.onlyHighestComboPoints = not self.db.profile.onlyHighestComboPoints
					end,
				},
				onlyHighestEnergy = {
					type = 'toggle',
					name = L["OnlyHighestEnergy"],
					desc = L["OnlyHighestEnergy_Desc"],
					get = function()
						return self.db.profile.onlyHighestEnergy
					end,
					set = function()
						self.db.profile.onlyHighestEnergy = not self.db.profile.onlyHighestEnergy
					end,
				},
			},
		}		
		-- Insert menu for each trigger
		for key, info in pairs(self.db.profile.combinations) do
			args[self.major .. "_Combination"].args[string.format("ComboPoint_%d-Energy_%d", info.comboPoints, info.energy)] = {
				type = 'group',
				name = string.format(L["ComboPointEnergyTitle"], info.comboPoints, info.energy),
				desc = L["ComboPointEnergyTitle_Desc"],
				args = { 
					list = {
						order = 1,
						type = 'group',
						name = coreL["AnimationList"],
						desc = coreL["AnimationList_Desc"],
						args = {
							new = Avion2Core:GetDefaultNewAnimationOption(info.animationList),
						},
					},
					cancelAnimation = {
						order = 2,
						type = 'select',
						name = coreL["CancelAnimation"],
						desc = coreL["CancelAnimation_Desc"],
						values = {
							[0] = coreL["DontCancel"],
							[1] = coreL["FadeOut"],
							[2] = coreL["Cancel"],
						},
						get = function()
							return info.cancelAnimation
						end,
						set = function(info, val)
							info.cancelAnimation = val
						end,
					},
					delete = {
						order = 3,
						type = 'execute',
						name = L["DeleteComboPointsEnergy"],
						desc = L["DeleteComboPointsEnergy_Desc"],
						func = function()
							Avion2Core:StopAnimationList(string.format("Avion2_ComboPointEnergy_%d-%d", info.comboPoints, info.energy), info.animationList, false)
							self.db.profile.combinations[key] = nil
							
							self:UpdateEventRegistration()
							Avion2Core:UpdateModuleOptionTable(self)
						end,
					},
				},
			}			
			-- Add options for each animation
			for animationListID, _ in pairs(info.animationList) do
				-- Combine all info to find this animation in module db in ONE string
				serializedInfo = self:Serialize(Avion2_CPE.major, "Combination", info.comboPoints, info.energy, animationListID)
				
				-- Update option table
				args[self.major .. "_Combination"].args[string.format("ComboPoint_%d-Energy_%d", info.comboPoints, info.energy)].args.list.args[string.format("Animation_%d", animationListID)] = Avion2Core:GetDefaultAnimationOption(info.animationList, animationListID, serializedInfo)
			end
		end

	-- Disabled, remove options
	else
		args[self.major .. "_Energy"] 		= nil
		args[self.major .. "_Combination"] 	= nil
	end
end


-------------------------------------------------------------------------------
-- Addon - Combo-Points - Events
-------------------------------------------------------------------------------
local lastComboPoints, bestComboPointHit, highestEnergyThreshold, highestComboPointThreshold
local runningCache = {}


-- Relay function
ComboPointRelay = function(eventName, unitID)
	if (unitID == "player" and UnitPowerType("player") == 3) then
		ComboPointUpdate()
		ComboPointEnergyUpdate()
		
		-- Update last combo-points
		Avion2_Energy.lastComboPoints 	= GetComboPoints()
	end
end
EnergyRelay = function(eventName, unitID)
	if (unitID == "player" and UnitPowerType("player") == 3) then
		EnergyUpdate()
		ComboPointEnergyUpdate()
	
		-- Update last energy
		Avion2_Energy.lastUnitPower 	= UnitPower("player")
	end
end


-- Energy changed
EnergyUpdate = function()
	-- Reset
	highestEnergyThreshold 	= -1
	
	-- For each threshold, trigger all animation since last update and cancel all old one
	for threshold, data in pairs(Avion2_CPE.db.profile.energy) do
		-- Check if Energy if greater than threshold
		if (Avion2_CPE.db.profile.onlyOneActiveEnergy or Avion2_CPE.db.profile.energy[threshold].cancelAnimation ~= 0 and UnitPower("player") < threshold) then -- and threshold < self.lastUnitPower
			Avion2Core:StopAnimationList(string.format("Avion2_EnergyThreshold_%d", index), Avion2_CPE.db.profile.energy[threshold].animationList, (not Avion2_CPE.db.profile.onlyOneActiveEnergy and Avion2_CPE.db.profile.energy[threshold].cancelAnimation == 1))
		end

		-- Check if Energy is inside interval [newEnergy - oldEnergy)
		if (Avion2_CPE.lastUnitPower < threshold and threshold <= UnitPower("player")) then
			-- If onlyOneActiveEnergy is set, keep track of loewst thresthold (if available) and don't fire an animation (yet)
			if (Avion2_CPE.db.profile.onlyOneActiveEnergy) then
				if (threshold > highestEnergyThreshold or highestEnergyThreshold == -1) then
					highestEnergyThreshold = threshold
				end
			else
				Avion2Core:TriggerAnimationList(string.format("Avion2_EnergyThreshold_%d", threshold), Avion2_CPE.db.profile.energy[threshold].animationList)
			end
		end
	end
	
	-- When onlyOneActiveEnergy is set, than highestEnergyThreshold is updated until highest threshold is found
	if (Avion2_CPE.db.profile.onlyOneActiveEnergy and Avion2_CPE.db.profile.energy[highestEnergyThreshold]) then
		Avion2Core:TriggerAnimationList(string.format("Avion2_EnergyThreshold_%d", threshold), Avion2_CPE.db.profile.energy[highestEnergyThreshold].animationList)
	end
end


-- Energy changed
ComboPointUpdate = function()
	-- Reset
	highestComboPointThreshold 	= -1
	
	-- For each threshold, trigger all animation since last update and cancel all old one
	for threshold, data in pairs(Avion2_CPE.db.profile.comboPoint) do
		-- Check if comboPoint if greater than threshold
		if (Avion2_CPE.db.profile.onlyOneActiveComboPoint or Avion2_CPE.db.profile.comboPoint[threshold].cancelAnimation ~= 0 and GetComboPoints() < threshold) then -- and threshold < self.lastUnitPower
			Avion2Core:StopAnimationList(string.format("Avion2_comboPointThreshold_%d", index), Avion2_CPE.db.profile.comboPoint[threshold].animationList, (not Avion2_CPE.db.profile.onlyOneActiveComboPoint and Avion2_CPE.db.profile.comboPoint[threshold].cancelAnimation == 1))
		end

		-- Check if comboPoint is inside interval [newcomboPoint - oldcomboPoint)
		if (Avion2_CPE.lastUnitPower < threshold and threshold <= GetComboPoints()) then
			-- If onlyOneActiveComboPoint is set, keep track of loewst thresthold (if available) and don't fire an animation (yet)
			if (Avion2_CPE.db.profile.onlyOneActiveComboPoint) then
				if (threshold > highestComboPointThreshold or highestComboPointThreshold == -1) then
					highestComboPointThreshold = threshold
				end
			else
				Avion2Core:TriggerAnimationList(string.format("Avion2_comboPointThreshold_%d", threshold), Avion2_CPE.db.profile.comboPoint[threshold].animationList)
			end
		end
	end
	
	-- When onlyOneActiveComboPoint is set, than highestComboPointThreshold is updated until highest threshold is found
	if (Avion2_CPE.db.profile.onlyOneActiveComboPoint and Avion2_CPE.db.profile.comboPoint[highestComboPointThreshold]) then
		Avion2Core:TriggerAnimationList(string.format("Avion2_comboPointThreshold_%d", threshold), Avion2_CPE.db.profile.comboPoint[highestComboPointThreshold].animationList)
	end
end


-- Combo point gained/spend or energy update
ComboPointEnergyUpdate = function()
	lastComboPoints 	= -1
	bestComboPointHit	= -1
	
	-- Iterate overl all combo-points
	for _, info in ipairs(Avion2_CPE.db.profile.combinations) do
		-- No need to iterate if we exactCP and tableCP < currentCP
		if (Avion2_CPE.db.profile.onlyHighestComboPoints and bestComboPointHit ~= -1 and bestComboPointHit ~= lastComboPoints) then
			break
		end
	
		-- Test if Combo-Points/Energy combination is found, that fits current values (and wasn't updated in the last push) ...
		if (GetComboPoints() 	>= info.comboPoints and info.comboPoints 	> Avion2_Energy.lastComboPoints				-- combopoints fits (not updated since last push)
		and UnitPower("player") >= info.energy 		and info.energy			> Avion2_Energy.lastUnitPower				-- energy fits (not updated since last push)
		and (not Avion2_CPE.db.profile.onlyHighestComboPoints or GetComboPoints() >= info.comboPoints)) then	-- 
			-- Need to know, when to cancel
			bestComboPointHit = info.comboPoints
			
			-- Only trigger animation if not onlyHighestEnergy or its the highestEnergy for this amount of cp
			if (not Avion2_CPE.db.profile.onlyHighestEnergy or lastComboPoints ~= GetComboPoints()) then
				Avion2Core:TriggerAnimationList(string.format("Avion2_ComboPointEnergy_%d-%d", info.comboPoints, info.energy), info.animationList)
			
				-- Track that we triggered this animation
				if (info.cancelAnimation ~= 0) then
					table.insert(runningCache, {comboPoints = info.comboPoints, energy = info.energy})
				end
			end
		end
		
		-- Need to keep track if iteration shows up lower combo-points
		lastComboPoints = info.comboPoints
	end

	-- Cancel all alnimation that need to be canceled
	for key, info in pairs(runningCache) do
		-- Need to be canceled? (Because not active anymore)
		if (GetComboPoints() < info.ComboPoints or UnitPower("player") < info.energy) then
			Avion2Core:StopAnimationList(string.format("Avion2_ComboPointEnergy_%d-%d", info.ComboPoints, info.energy), info.animationList, (info.cancelAnimation == 1))
			
			-- Delete from track-list
			table.remove(runningCache, key)
		end
	end
end
